# -*- coding: utf-8 -*-
"""CLI interface to wbuild."""

import sys
import os
import click
import wbuild
import pathlib
import shutil
import distutils.dir_util
import click_log
import logging
import wbuild.utils as utils
logger = logging.getLogger(__name__)
click_log.basic_config(logger)

def setup_paths():
    """Setup the wbuild paths
    """
    templatePath = pathlib.Path(wbuild.__file__).parent / 'template'
    wbuildPath = pathlib.Path(wbuild.__file__).parent / '.wBuild'
    demoPath = pathlib.Path(wbuild.__file__).parent / 'demo'
    return templatePath, wbuildPath, demoPath

@click.group()
@click_log.simple_verbosity_option(logger)
@click.version_option('1.7.0',prog_name='wBuild')
def main():
    pass


# --------------------------------------------
# commands

@main.command()
def init():
    """Initialize the repository with wbuild.

    This will prepare wBuild in the current project
    """
    if os.path.exists(".wBuild"):
        logger.error("ERROR: .wBuild already exists. Use update if you want to update the version")
        sys.exit(2)
    templatePath, wbuildPath, demoPath = setup_paths()
    distutils.dir_util.copy_tree(str(wbuildPath), './.wBuild')
    if not os.path.isfile("Snakefile"):
        shutil.copy(str(templatePath / 'Snakefile'), '.')
    if not os.path.isfile("wbuild.yaml"):
        shutil.copy(str(templatePath / 'wbuild.yaml'), '.')
    
    ### search for a file containing the word readme and .md
    readme_exists = False
    onlyfiles = [f for f in os.listdir(".") if os.path.isfile(os.path.join(".", f))]
    for f in onlyfiles:
        if ("readme" in f) and f.endswith(".md"):
            readme_exists = True
    if not readme_exists:
        copyReadme = input("wBuild needs readme.md in a root folder of your project. Shall we create the default "
                           "one? (y/n)")
        if 'y' in copyReadme:
            shutil.copy(str(templatePath / 'readme.md'), '.')
    utils.writeWbuildVersion()

    logger.info("init...done")


@main.command()
def demo():
    """Setup a demo wBuild demo project
    """
    if os.path.exists(".wBuild"):
        logger.error("ERROR: .wBuild already exists. Run demo in empty folder.")
        sys.exit(2)
    templatePath, wbuildPath, demoPath = setup_paths()
    shutil.copy(str(templatePath / 'Snakefile'), '.')
    shutil.copy(str(templatePath / 'wbuild.yaml'), '.')
    distutils.dir_util.copy_tree(str(wbuildPath), './.wBuild')
    distutils.dir_util.copy_tree(str(demoPath), '.')
    utils.writeWbuildVersion()
    logger.info("demo...done")


@main.command()
def update():
    """Update the .wBuild folder to the most recent version of wBuild
    """
    # - check if .wBuild exists
    #   - if it doesn't, return the error
    # - if it does, update it
    if not os.path.exists(".wBuild"):
        raise ValueError(".wBuild doesn't exists. Please run wBuild init first or move to the right directory")
    logger.info("Removing .wBuild")
    shutil.rmtree("./.wBuild")
    # import subprocess
    # deprecatedPackages = subprocess.check_output(['pip','list','--outdated']).decode("utf-8")
    # if 'wbuild' in deprecatedPackages:
    #     logger.warning("Newer version of wBuild available.")
    #     updateConf = input("Update wBuild using pip (requires internet connection)? (y/n)")
    #     if 'y' in updateConf:
    #         subprocess.call(['pip','install','wbuild','--upgrade'])
    #         logger.info("wBuild successfully updated!")
    logger.info("Running .init")
    templatePath, wbuildPath, demoPath = setup_paths()
    distutils.dir_util.copy_tree(str(wbuildPath), './.wBuild')
    utils.writeWbuildVersion()
    logger.info("update...done")



import fnmatch
import os
import sys
import yaml
import yaml.scanner
import yaml.parser
import yaml.error
import operator
import re
from functools import reduce
from snakemake.logging import logger
from snakemake import get_argument_parser, parse_config, SNAKEFILE_CHOICES

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def checkFilename(filename):
    """
    :param filename: to check
    :return: has appropriate name?
    :raises: ValueError if the name is inappropriate
    """
    if " " in filename:
        raise ValueError("Spaces are not allowed in the filenames. File: {0}",filename)
    if "-" in os.path.basename(filename):
        raise ValueError("- are not allowed in the filenames. File: {0}", filename)
    return True

def findFilesRecursive(startingPath, patterns):
    """
    :param startingPath: root path of the search
    :param patterns: patterns to search file names for
    :return: paths to files matching the patterns
    """
    matchedFilepaths = []
    for root, dirnames, filenames in os.walk(startingPath):
        dirnames[:] = [d for d in dirnames if not d[0] == '_']
        dirnames[:] = [d for d in dirnames if not d[0] == '.']
        for file in reduce(operator.add, (fnmatch.filter(filenames, p) for p in patterns)):
            checkFilename(file)
            absFilepath = os.path.join(root, file)
            if not absFilepath in matchedFilepaths:
                matchedFilepaths.append(absFilepath)
    sortedMatchedFilepaths = sorted(matchedFilepaths)
    conf = Config()
    regex = re.compile(conf.get("fileRegex"))
    reFiles = list(filter(regex.search, sortedMatchedFilepaths))
    logger.debug("Found files in scope of wBuild: " + str(reFiles) + ".\n")
    return reFiles


def parseYAMLHeader(filepath):
    """

    :param filepath: path to the file
    :return: String representation of the YAML header in the file, including inter-document framing ("---")
    """
    yamlHeader = []
    for i, line in enumerate(open(filepath).readlines()):
        # process
        yamlHeader.append(line.strip()[2:])

        # terminate if that's already "#'---" (=end of YAML-designated area)
        if i != 0 and line.startswith("#'---"):
            break

    result = '\n'.join(yamlHeader)
    logger.debug("Got " + result + "as a result of parsing YAML header from " + filepath + ".\n")
    return result


def hasYAMLHeader(filepath):
    """
    :param filepath: path to the file
    :return: file contains YAML header?
    """
    with open(filepath, "r") as f:
        lines = f.readlines()
    line = lines[0]
    if(line.startswith("#'---")):
        return True
    return False


def parseWBInfosFromRFiles(script_dir="Scripts", htmlPath="Output/html"):
    """

    :param script_dir: Relative path to the Scripts directory
    :param htmlPath: Relative path to the html output path
    :return: a list of dictionaries with fields:
      - file - what is the input R file
      - outputFile - there to put the output html file
      - param - parsed yaml params
    """
    parsedInfos = []
    #errorOccured = False
    for filename in findFilesRecursive(script_dir, ['*.r', '*.R']):
        if not hasYAMLHeader(filename):
            # Ignore files without YAML infos
            continue
        header = parseYAMLHeader(filename)
        # run all the synthax checks - will raise an error if it fails
        yamlParamsDict = parseYamlParams(header, filename)
        if yamlParamsDict == None: #parsing error occured
            continue #go on parsing next file


        if type(yamlParamsDict) is str: #allow parsing one tag without double points as string; put it in a dict and check later on
            yamlParamsDict = {yamlParamsDict: None}

        if('wb' in yamlParamsDict):# the header contains wb informations
            outFile = htmlPath + "/" + pathsepsToUnderscore(os.path.splitext(filename)[0]) + ".html"
            parsedInfos.append({'file': linuxify(filename), 'outputFile': outFile, 'param': yamlParamsDict})

    logger.debug("Parsed informations from R files: " + str(parsedInfos))
    #if errorOccured:
    #    raise ValueError("Errors occured in parsing the R files. Please fix them.") TODO really raise a ValueError?
    return parsedInfos

def parseWBInfosFromRFile(filename, htmlPath="Output/html"):
    """
    :param filename: Relative path to the Scripts directory
    :param htmlPath: Relative path to the html output path
    :return: a list of dictionaries with fields:
      - filen - what is the input R file
      - outputFile - there to put the output html file
      - param - parsed yaml params
    """
    parsedInfos = []
    #errorOccured = False
    if not hasYAMLHeader(filename):
        # Ignore files without YAML infos
        print('Header not valid')
    header = parseYAMLHeader(filename)
    # run all the synthax checks - will raise an error if it fails
    yamlParamsDict = parseYamlParams(header, filename)
    if type(yamlParamsDict) is str: #allow parsing one tag without double points as string; put it in a dict and check later on
        yamlParamsDict = {yamlParamsDict: None}
    if('wb' in yamlParamsDict):# the header contains wb informations
        outFile = htmlPath + "/" + pathsepsToUnderscore(os.path.splitext(filename)[0]) + ".html"
        parsedInfos.append({'file': linuxify(filename), 'outputFile': outFile, 'param': yamlParamsDict})

    logger.debug("Parsed informations from R files: " + str(parsedInfos))
    #if errorOccured:
    #    raise ValueError("Errors occured in parsing the R files. Please fix them.") TODO really raise a ValueError?
    return parsedInfos


def parseMDFiles(script_dir="Scripts", htmlPath="Output/html"):
    """

    :param script_dir: Relative path to the Scripts directory
    :param htmlPath: Relative path to the html output path
    :return: a list of dictionaries with fields:
      - file - what is the input .md file
      - outputFile - there to put the output html file
      - param - parsed yaml header - always an empty list
    """
    logger.debug("Finding .md files:\n")
    foundMDFiles = []
    for f in findFilesRecursive(script_dir, ['*.md']):
        outFile = htmlPath + "/" + pathsepsToUnderscore(os.path.splitext(f)[0])+ ".html"
        logger.debug("Found " + outFile + ".\n")
        foundMDFiles.append({'file': linuxify(f), 'outputFile': outFile, 'param': []})
    return foundMDFiles


def getYamlParam(r, paramName):
    if 'wb' in r['param'] and type(r['param']['wb']) is dict and paramName in r['param']['wb']:
        foundParam = r['param']['wb'][paramName]
        return foundParam
    return None

def parseYamlParams(header, f):
    """
    :param header: String form of YAML header
    :param f: Filename of a file from where the header was parsed
    :return: Parameters dictionary parsed from the header; None if parsing errors occured
    """
    try:
        param = next(yaml.safe_load_all(header))
    except (yaml.scanner.ScannerError, yaml.parser.ParserError, yaml.error.YAMLError, yaml.error.MarkedYAMLError) as e:
        if hasattr(e, 'problem_mark'):
            if e.context != None:
                logger.error('Error while parsing YAML area in the file ' + f + ':\n' + str(e.problem_mark) + '\n  ' +
                      str(e.problem) + ' ' + str(e.context) +
                      '\nPlease correct the header and retry.')
            else:
                logger.error('Error while parsing YAML area in the file ' + f + ':\n' + str(e.problem_mark) + '\n  ' +
                      str(e.problem) + '\nPlease correct the header and retry.')
        else:
            logger.error("YAMLError parsing yaml file.")

        return None
    except Exception as e:
        print(bcolors.FAIL + bcolors.BOLD + 'Could not parse', f,
              '. Include valid yaml header. Not showing any further errors. \n',
              'Errors {0}'.format(e) + bcolors.ENDC)
        return None

    logger.debug("Parsed params: " + str(param) + "\n.")
    return param

def pathsepsToUnderscore(systemPath, dotsToUnderscore = False, trimPrefix=True):
    """
    Convert all system path separators and dots to underscores. Product is used as a unique ID for rules in scanFiles.py or the output HTML files
    :param systemPath: path to convert in
    :param dotsToUnderscore: if the dot should be converted as well. Defaults to false
    :return: path string with converted separators
    """
    if trimPrefix:
        conf = Config()
        systemPath = removeFilePrefix(systemPath, conf.snakeroot)

    if dotsToUnderscore:
        return systemPath.replace('.', '_').replace('/', '_').replace('\\', '_')
    return systemPath.replace('/', '_').replace('\\', '_')

def linuxify(winSepStr, doubleBackslash = False):
    """
    Convert windows (path) string to the linux format.

    :param winSepStr: (path) string with windows-like "\" separators
    :param doubleBackslash: if the slashes in the winSepStr are double (happens when you read a macro string raw. Ex.: "C:\\Program Files\\a.txt"
    :return: str with substituted "\" -> "/"
    """
    if doubleBackslash:
        return winSepStr.replace("\\\\", "/")
    return winSepStr.replace("\\", "/")

class Config:

    sysargs = None
    args = None
    config = None
    config_dict = None
    path = "wBuild.yaml"
    snakefile = "Snakefile"
    snakeroot = ""
    instance = None
    


    def __init__(self):
        # check if it is already initialized
        if self.instance != None:
            self.config = self.instance.config
            self.conf_dict = self.instance.conf_dict
            self.args = self.instance.args
            self.path = self.instance.path
            self.snakefile = self.instance.snakefile
            self.snakeroot = self.instance.snakeroot
            return

        # we dont need the first argument aka call to snakemake
        self.sysargs = sys.argv[1:]

        parser = get_argument_parser()
        self.args = parser.parse_args(self.sysargs)
        self.path = self.args.configfile
        self.snakefile = self.args.snakefile
        self.config = parse_config(self.args)

                
        if self.path is None:
            for p in ["wbuild.yaml", "config.yaml", "wBuild.yaml"]:
                if os.path.exists(p):
                    self.path = p
                    break
        else:
            if type(self.path) is list:
                if len(self.path) != 1:
                    raise Exception("Path is a list of more than one element! '" + str(self.path) + "'")
                self.path = self.path[0]
            self.path=os.path.abspath(self.path)

        # this is taken from the snakemake main file
        if self.snakefile is None:
            for p in SNAKEFILE_CHOICES:
                if os.path.exists(p):
                    self.snakefile = p
                    break
        self.snakeroot = os.path.dirname(self.snakefile)

        #load defaults
        self.loadDefaultConfiguration()

        try:
            fh = open(self.path, "r")
        except IOError:
            raise IOError("Can not read config. Are you sure you have enough "
                          "rights and config path (wbuild.yaml) is right?")
        configDict = next(yaml.safe_load_all(fh))
        if configDict == None:
            logger.error("Error parsing wbuild.yaml - format is wrong. Working with defaults...")
        else:
            self.conf_dict = merge_two_dicts(self.conf_dict, configDict)
        #fill Singleton
        Config.instance = self

    def loadDefaultConfiguration(self):
        abspathSnakefile = os.path.abspath(self.snakefile)
        prefixScripts = self.snakeroot
        if len(prefixScripts) > 0:
            prefixScripts = prefixScripts + "/"

        self.conf_dict = {"htmlOutputPath": "Output/html",
                          "processedDataPath": "Output/ProcessedData",
                          "scriptsPath": prefixScripts + "Scripts",
                          "projectTitle": "Project"}

    def get(self, attrname):
        if (attrname in self.conf_dict):
            return self.conf_dict[attrname]
        else:
            raise AttributeError("There is no attribute " + attrname +
                                 " in the configuration file loaded!")

def merge_two_dicts(x, y):
    z = x.copy()  # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z


def writeWbuildVersion():
    """
    Write wBuild version to .wBuild/.version
    """
    with open(".wBuild/.version", 'w') as file:
        import wbuild
        file.write(wbuild.__version__)
        file.close()

def wbuildVersionIsCurrent():
    """
    Read wBuild version from .wBuild/.version and compare it to wbuild module version from pckg mngr.
    :return: True if wBuild up-to-date, False if not
    """
    with open(".wBuild/.version", 'r') as file:
        static_v = file.read()

    import wbuild
    dynamic_v = wbuild.__version__
    return dynamic_v in static_v

def removeFilePrefix(f, prefix):
    if f.startswith(prefix):
        f = f[len(prefix):]
        if len(prefix) > 0 and f.startswith("/"):
            f = f[1:]

    absPrefix = os.path.abspath(prefix)
    if f.startswith(absPrefix):
        f = f[(len(absPrefix)+1):]
    return(f)







import os
import sys
import pathlib
import shutil
import wbuild

from string import Template
from os import listdir
from os.path import isfile, join
from wbuild.utils import parseWBInfosFromRFiles, parseMDFiles, \
        getYamlParam, Config, removeFilePrefix

sys.path.insert(0, os.getcwd() + "/.wBuild")  # TODO - is this line required?

def writeSubMenu(top, wbData, level):
    """
    Recursive call to construct the dropdown list and hover-over side-menus in it adhereing to a "top" toolbar category.

    :param top: "top" toolbar directory to be appointed to
    :param wbData: wb relevant data of all scanned files
    :param level: deepness of the current submenu (first dropdown list, then hover-over side-menus in the html)
    :return: deeply constructed dropdown list of the top toolbar category as an HTML string
    """
    
    menuString = ''
    temp = []
    newWb = []
    for r in wbData:
        temptemp = pathlib.PurePath(r['file']).parts[level - 1]
        if (pathlib.PurePath(r['file']).parts[level - 1] == top):
            # Is it a file
            if(len(pathlib.PurePath(r['file']).parts) == (level + 1)):
                if getYamlParam(r, 'type') != 'script' and getYamlParam(r, 'type') != 'noindex':
                    menuString += ('<li><a href="javascript:navigate(\'' +
                            pathlib.PurePath(r['outputFile']).name + '\');">' +
                            pathlib.PurePath(r['file']).parts[level] + '</a></li>\n')
                continue
            temp.append(pathlib.PurePath(r['file']).parts[level])
            newWb.append(r)
    
    temp = sorted(set(temp))
    for top in temp:
        menuString += '<li class="dropdown-submenu">\n'
        menuString += '    <a tabindex="-1" href="#">' + top + '</a>\n'
        menuString += '    <ul class="dropdown-menu">\n'
        menuString += '    ' + writeSubMenu(top, newWb, level + 1)
        menuString += '    </ul>\n'
        menuString += '</li>\n'
    return menuString


def getRecentMenu():
    """
    Support recently edited files list to the HTML web output.

    :return: HTML string: "Recently viewed" menu contents
    """
    conf = Config()
    htmlOutputPath = conf.get("htmlOutputPath")
    rFiles = sorted([join(htmlOutputPath, f) for f in listdir(htmlOutputPath)
        if isfile(join(htmlOutputPath, f))], key=os.path.getmtime, reverse=True)
        
    ## delete all files containing the word "index from html menu "
    rFiles = [f for f in rFiles if "index" not in f]
    rFiles = rFiles[:10]

    # open recent Files in new tab 
    menuString = ""
    for f in rFiles:
        fo = pathlib.PurePath(f).name
        
        # Open in a new tab
        #menuString += ('<p><a href='+ fo + ' target="_blank">' + fo.replace('_', ' ').replace('.html', '') +
        #        '</a></p>\n')
        
        # Open in same tab
        menuString += ('<p><a href="javascript:navigate(\'' +
                fo + '\');" >' + fo.replace('_', ' ').replace('.html', '') +
                '</a></p>\n')
    return menuString

def writeReadme():
    """ Extract readme file from readme path in config. 
    If not specified file containing <readme> in scriptsPath with be chosen"""
      
    conf = Config()
    scriptsPath = conf.get("scriptsPath")
    snakeroot = conf.snakeroot
    
    try:
        filename_readme = conf.get("readmePath") #### should be .md file
    except AttributeError as e:
        filename_readme = ""
        onlyfiles = [f for f in os.listdir(snakeroot) if os.path.isfile(os.path.join(snakeroot, f))]
        for f in onlyfiles:
            if ("readme" in f) and f.endswith(".md"):
                filename_readme = f
    
    filename_readme = filename_readme.replace(".md", ".html")
    readmeString = '<li><a href="javascript:navigate(' + " '{}'".format(filename_readme) + ');">Readme</a></li> '
    readmeIframeString = '<iframe id="Iframe" src="' + filename_readme + '" width=100% height=95% ></iframe> '
    readmeFilename = ' "{}" '.format(filename_readme)
    
    return readmeString, readmeIframeString, readmeFilename           
        
def writeDepSVG():
    """ Search for rule graph. If path not specified in config, take default dep.svg in snakeroot path"""
    conf = Config()
    scriptsPath = conf.get("scriptsPath")
    htmlOutputPath = conf.get("htmlOutputPath")
    snakeroot = conf.snakeroot
    foldername = snakeroot.split("/")[-1]
    
    try:
        filename_SVG = conf.get("ruleGraphPath") #### should be .md file
    except AttributeError as e:
        ### try with default name "dep.svg"
        if os.path.isfile(os.path.join(htmlOutputPath, "dep.svg")):
            filename_SVG = "dep.svg"  
        else:
            ### search for files containing "svg" and foldername 
            filename_SVG = ""
            onlyfiles = [f for f in os.listdir(htmlOutputPath) if os.path.isfile(os.path.join(htmlOutputPath, f))]
            for f in onlyfiles:
                if (foldername in f) and f.endswith(".svg"):
                    filename_SVG = f
 
    svgString = '<li><a href="javascript:navigate(' + "'{}'".format(filename_SVG) + ');">Dependency</a></li>' 
    return svgString

def writeIndexHTMLMenu():
    """
    Scan for files involved in the current HTML rendering and fill the HTML quick access toolbar correspondingly
    """
    conf = Config()
    
    htmlOutputPath = conf.get("htmlOutputPath")
    scriptsPath = conf.get("scriptsPath")
    pageTitle = conf.get("projectTitle")
    snakeroot = conf.snakeroot

    wbData = parseWBInfosFromRFiles(script_dir=scriptsPath, htmlPath=htmlOutputPath)
    mdData = parseMDFiles(script_dir=scriptsPath, htmlPath=htmlOutputPath)
    wbData += mdData
    temp = []
    
    # for all of the scanned files, collect their paths
    for r in wbData:
        # this is needed so the relative path to "../wbuild/Snakefile" is not
        # part of the html sub menu
        r['file'] = removeFilePrefix(r['file'], snakeroot)
        temp.append(pathlib.PurePath(r['file']).parts[1])

    menuString = ""
    for top in sorted(set(temp)):
        menuString += (
            '<li class="dropdown">\n' +
            #write the current directory's name to the main ("top") toolbar tab
            '   <a href="#" class="dropdown-toggle" data-toggle="dropdown" ' +
                    'role="button" aria-haspopup="true" aria-expanded="false">' +
                    top + '<span class="caret"></span></a>\n'
            '   <ul class="dropdown-menu multi-level" role="menu">\n' +
            #write sub-directories to the dropdown list of the "top" tabs
            writeSubMenu(top, wbData, 2) +
            '   </ul>\n' +
            '</li>\n')
            
    readmeString, readmeIframeString, readmeFilename = writeReadme()
    depSVGString = writeDepSVG()
    
    #fill the HTML template with the constructed tag structure
    wbuildPath = pathlib.Path(wbuild.__file__).parent
    
    template = open(str(wbuildPath / "html/template.html")).read()
    template = Template(template).substitute(menu=menuString, title=pageTitle, rf=getRecentMenu(),
                        readme=readmeString, readmeIframe=readmeIframeString, readmeFilename=readmeFilename
                        , depSVG=depSVGString)
                        
                        
    try:
        filename_index = conf.get("htmlIndex")
    except AttributeError as e:
        filename_index = "index.html"
    
    try: 
        indexWithFolderName = conf.get("indexWithFolderName")
    except:
        indexWithFolderName = False
    
    if indexWithFolderName:
        abs_path = str(os.path.abspath(scriptsPath))
        name = abs_path.split("/")[-2]
        filename_index = name + "_" + filename_index
    
    f = open(htmlOutputPath + '/' + filename_index, 'w')
    f.write(template)
    f.close()


def ci():
    writeIndexHTMLMenu()

    conf = Config()
    htmlOutputPath = conf.get("htmlOutputPath")
    libDir = htmlOutputPath + "/lib"

    if os.path.exists(libDir):
        shutil.rmtree(libDir)
    
    wbuildPath = pathlib.Path(wbuild.__file__).parent
    shutil.copytree(str(wbuildPath) + "/html/lib", libDir)
    


import os
import sys
import pathlib
import re
from snakemake.logging import logger
from wbuild.utils import parseWBInfosFromRFiles, parseWBInfosFromRFile, parseMDFiles, getYamlParam, pathsepsToUnderscore, \
    Config, wbuildVersionIsCurrent, bcolors
import tempfile
import wbuild

pathsep = "/"
sys.path.insert(0, os.getcwd() + "/.wBuild")

#wbuildPath = pathlib.Path(wbuild.__file__).parent
#Config().conf_dict.setdefault('wBuildPath',str(wbuildPath))
#print(Config().conf_dict.items())

# SNAKEMAKE  = ["input", "output", "threads"]

#dict containing snakemake supported fields
SNAKEMAKE_FIELDS = ["input",
                    "output",
                    "params",
                    "threads",
                    "resources",
                    "priority",
                    "version",
                    "log",
                    "message",
                    "run",
                    "shell",
                    "script"]


def writeDependencyFile():
    """
    Entry point for writing .wBuild.depend.
    """
    #if not wbuildVersionIsCurrent():
    #    print(bcolors.WARNING + "Version of the project's static .wBuild lib is not the same as the dynamically loaded "
    #                            "wBuild"
    #                            "version. It is strongly recommended to update .wBuild lib using \'wbuild update\'; "
    #                            "otherwise, the consistency of the build can not be guaranteed." + bcolors.ENDC)
    logger.info("Structuring dependencies...")
    conf = Config()
    htmlOutputPath = conf.get("htmlOutputPath")
    logger.debug("Loaded config.\n html output path (key htmlOutputPath): " + htmlOutputPath + "\n")
    scriptsPath = conf.get("scriptsPath")
    wbData = parseWBInfosFromRFiles(script_dir=scriptsPath, htmlPath=htmlOutputPath)
    mdData = parseMDFiles(script_dir=scriptsPath, htmlPath=htmlOutputPath)
    dependFile = tempfile.NamedTemporaryFile('w',delete=False)
    with dependFile as f: #start off with the header
        f.write('######\n')
        f.write('#This is a autogenerated snakemake file by wBuild\n')
        f.write('#wBuild by Leonhard Wachutka\n')
        f.write('######\n')
        # write rules
        for r in wbData:
            writeRule(r, f)
        # write md rules
        for r in mdData:
            writeMdRule(r, f)

        # write build index rule
        writeIndexRule(wbData, mdData, f)
    logger.info("Dependencies file generated.\n")

    return(dependFile.name)

def writeWBParseDependencyFile(filename):
    """
    Entry point for writing .wBuild.depend. for the wbParseFunction in R
    """
    conf = Config()
    htmlOutputPath = conf.get("htmlOutputPath")
    wbData = parseWBInfosFromRFile(filename=filename, htmlPath=htmlOutputPath)
    
    with open('.wBuild.depend', 'w') as f: #start off with the header
        f.write('######\n')
        f.write('#This is a autogenerated snakemake file by wBuild\n')
        f.write('#wBuild by Leonhard Wachutka\n')
        f.write('######\n')
        # write rules
        for r in wbData:
            writeRule(r, f, True)
        writeIndexRule(wbData, list(), f)
    logger.info("Dependencies file generated.\n")

def joinEmpty(string_list):
    """
    :param string_list:
    :return: sting representation of a list without the blank elements.
    """
    return ", ".join([x for x in string_list if x.strip() != ''])


def escapeSMString(item):
    """
    Convert item to the appropriate string representation.

    :param item: string or dict
    :return: "key = 'value'" (dict), "'value'" (string) or '' (other type)
    """
    if type(item) is dict:
        #return "key = value"
        return str(list(item.keys())[0]) + ' = ' + escapeSMString(str(list(item.values())[0]))
    elif type(item) is str:
        if item.startswith("`sm ") and item.endswith("`"): #strip `sm ... `
            return item[4:-1]
        return "'" + item + "'" #return item as quoted string
    return ''


def ensureString(elem):
    if elem is None:
        return ''
    elif type(elem) is list:
        if len(elem) == 0:
            return ''
        else:
            # make sure each element is a character
            elem = [escapeSMString(item) for item in elem]
            elem = [x for x in elem if str(x).strip() != '']

            return ", ".join(elem)
    elif type(elem) is str:
        if "," not in elem:
            return "'" + elem + "'"
        else:
            return elem
    elif type(elem) is int:
        return str(elem)
    elif type(elem) is dict:
        elemArr = [k + " = " + escapeSMString(elem[k]) for k in elem]
        return ", ".join(elemArr)
        #raise TypeError("A wBuild tag is a dict, whereas it can be list or string only. Please check if you have indented"
        #                " all the YAML \'output\' tags in your scripts properly. \nwBuild output should be listed one level deeper than external (knitr) output."
        #                "(The latter one should be on the same level with \'wb\' tag)")
    else:
        raise TypeError("Can't parse type " + str(type(elem)) + "as a valid workflow information under a wBuild YAML tag (input or output).")


def dumpSMRule(ruleInfos, outputFile, inputFile):
    """
    Write the rule to the file.

    :param ruleInfos: dictionary containing all the rule's data
    :param outputFile: file to print the rule to
    :param inputFile: the object of the rule
    """
    if 'py' in ruleInfos:
        code = ruleInfos['py']
        if type(code) is str:
            outputFile.write(insertPlaceholders(code, inputFile))
        elif type(code) is list:
            [outputFile.write(insertPlaceholders(line, inputFile) + '\n') for line in code]

    outputFile.write('rule ' + ruleInfos['rule'] + ':\n')
    for field in SNAKEMAKE_FIELDS:
        if field in ruleInfos:
            outputFile.write('    ' + field + ': ' + str(ruleInfos[field]) + '\n')


def insertPlaceholders(dest, source):
    """
    Infer placeholders' substitutions.

    :param dest: string to replace placeholders in
    :param source: file; from its path we infer the placeholders values
    :return: dest with replaced placeholders
    """
    path = pathlib.Path(source) #get the path to the file
    conf = Config()
    processedDataPath = conf.get("processedDataPath")
    PD = pathlib.Path(processedDataPath)

    PP = path.parts[-2]
    dest = dest.replace("{wbPD}", str(PD))
    dest = dest.replace("{wbPP}", str(PP))

    if len(path.parts) <= 2 and bool(re.search('{wbP(D_P*)?}', source)):
        print("If using placeholders please make sure you have the right",
              " directory structure.")

    if len(path.parts) > 2:
        P = path.parts[-3]
        dest = dest.replace("{wbPD_P}", str(PD / P))
        dest = dest.replace("{wbPD_PP}", str(PD / P / PP))
        dest = dest.replace("{wbP}", str(P))

    return dest


def writeRule(r, file, dump=False):
    """
    Write Snakemake rule from the parsed WB header informations.

    :param r: parsed WB data dictionary entry
    :param file: to write the rule to
    """
    #TODO cleanup boilerplate commented code here
    wbInfos = r["param"]["wb"]
    inputFile = r['file'] #input R script for Snakemake

    wbuildPath = pathlib.Path(wbuild.__file__).parent
    # extract rule
    # rule = r['file'].replace('.','_').replace('/','_')

    if wbInfos == None:
        return

    # determine input, output and script
    wbInfos["input"] = insertPlaceholders(joinEmpty([ensureString(wbInfos.get("input")), "RScript = '" + inputFile + "'"]), inputFile)
    if wbInfos.get("type") == 'script':
        wbInfos["output"] = insertPlaceholders(ensureString(wbInfos.get("output")), inputFile)
        wbInfos["script"] = "'" + os.getcwd() + '/' + inputFile + "'"
    elif wbInfos.get("type") == 'noindex':
        wbInfos["output"] = insertPlaceholders(ensureString(wbInfos.get("output")), inputFile)
        wbInfos["script"] = "'" + str(wbuildPath / 'R'/'wBRender.R') + "'"
    else:
        wbInfos["output"] = insertPlaceholders(joinEmpty([ensureString(wbInfos.get("output")), "wBhtml = '" + r['outputFile'] + "'"]), inputFile)
        wbInfos["script"] = "'" + str(wbuildPath / 'R'/'wBRender.R') + "'"
    if dump==True:
        wbInfos["script"] = "'" + str(wbuildPath / 'R'/'wBSMDump.R') + "'"
    
    for i in SNAKEMAKE_FIELDS:
        if i not in ['output', 'script', 'input', 'run', 'shell'] and i in wbInfos.keys():
            wbInfos[i] = ensureString(wbInfos.get(i))
    # remove wb related elements
    # wbInfos = {key: wbInfos[key] for key in wbInfos if key not in WB_FIELDS}
    # if not set(wbInfos.keys()).issubset(SNAKEMAKE_FIELDS):
    #    Warning("File: {0}. The following fields don't correspond to any snakemake or wBuild tag: {1}"
    #            .format(r['file'], ",".join(set(wbInfos.keys()).difference(SNAKEMAKE_FIELDS))))

    # remove fields not in SNAKEMAKE_FIELDS
    # wbInfos = {key: wbInfos[key] for key in wbInfos if key in SNAKEMAKE_FIELDS}
    wbInfos['rule'] = pathsepsToUnderscore(r['file'], True) # convert filepath to the unique id of the rule
    # write to file
    file.write('\n')
    # dumpDict = {'rule ' + rule: wbInfos}
    # file.write(yaml.dump(dumpDict, default_flow_style = False, indent=4).replace("\'\'\'", "'").replace("\'\'", "'"))
    dumpSMRule(wbInfos, file, inputFile)
    file.write('\n')


def writeMdRule(ruleInfos, file):
    """
    :param ruleInfos:
    :param file: file to write the rule to
    """
    file.write('\n')
    file.write('rule ' + pathsepsToUnderscore(ruleInfos['file'], True) + ':\n')
    file.write('    input: "' + ruleInfos['file'] + '"\n')
    file.write('    output: "' + ruleInfos['outputFile'] + '"\n')
    file.write('    shell: "pandoc --from markdown --to html --css {config[wBuildPath]}/html/lib/github.css --toc --self-contained -s -o {output} {input}"\n')

    file.write('\n')


def writeIndexRule(wbRRows, wbMDrows, file):
    """
    Write the rule of mapping the R and md wbData to the index.html.

    :param wbRRows: info dict parsed from R wB files
    :param wbMDrows: info dict parsed from MD wB files
    :param file: file to print the index rule to
    """
    inputFiles = []
    for r in wbRRows:
        #ignore if the file is script or noindex
        if getYamlParam(r, 'type') == 'script' or getYamlParam(r, 'type') == 'noindex':
            continue

        inputFiles.append(r['outputFile'])

    for r in wbMDrows:
        inputFiles.append(r['outputFile'])

    conf = Config()
    htmlOutputPath = conf.get("htmlOutputPath")
    try:
        filename_index = conf.get("htmlIndex")
    except AttributeError as e:
        filename_index = "index.html"
    try: 
        indexWithFolderName = conf.get("indexWithFolderName")
    except:
        indexWithFolderName = False

    if indexWithFolderName:
        scriptsPath = conf.get("scriptsPath")
        abs_path = str(os.path.abspath(scriptsPath))
        name = abs_path.split("/")[-2]
        filename_index = name + "_" + filename_index
    

    file.write('\n')
    file.write('rule Index:\n')
    file.write('    input: \n        "' + '",\n        "'.join(inputFiles) + '"\n')
    file.write('    output: "' + htmlOutputPath + '/' + filename_index + '" \n')
    # file.write('    script: ".wBuild/createIndex.py"\n')
    file.write('    run:\n')
    file.write('        import wbuild.createIndex\n')
    file.write('        wbuild.createIndex.ci()\n')

    file.write('\n')


if __name__ == "__main__":
    writeDependencyFile()
